% Construct MDP for inverted pendulum problem.
% 
% Arguments:
%
% h          time step
% v_max      maximum velocity
% u_max      maximum control magnitude
% n1         # of grid points on state x1 (arc) dimension
% n2         # of grid points on state x2 (velocity) dimension
% nu         # of grid points on u (control) dimension
% a          models gravity, mass, and length ( > 0)
% b          damping factor ( > 0)
% k          sharpness of cost function ( > 0 )
% r          weight of control cost ( > 0 )
function [y1_mat, y2_mat, l_mat] = construct_mdp(h, v_max, u_max, n1, n2, nu, a, b, k, r)
	% Use number of grid points to discretize dimensions
    
    x1_vals = linspace(0, 2*pi, n1); % divide full arc into n1 pts
    x2_vals = linspace(-v_max, v_max, n2); % divide velocity into n2 pts
    u_vals = linspace(-u_max, u_max, nu); % divide velocity into nu pts

%         
%         disp('h:');
%         disp(h);
%         
%         disp('x1:');
%         disp(x1_vals);
%         
%         disp('x2:');
%         disp(x2_vals);
%         
%         disp('u:');
%         disp(u_vals);
%    

    % y1_mat and y2_mat are 3D arrays mapping states & input to outputs;
    % the point (x1_idx, x2_idx, u_idx) will map to the corresponding y1
    % or y2 for x1, x2 and u. Note that the arrays store structs where
    % the neighboring states and respective transition probabilities are
    % stored, since the exact y1 and y2 values will usually not map to
    % actual states in the state space.
    y1_mat(1:n1, 1:n2, 1:nu) = struct('points',[],'indices',[],'probs',[],'input',[]);
    y2_mat(1:n1, 1:n2, 1:nu) = struct('points',[],'indices',[],'probs',[],'input',[]);
    % l_mat is a similar 3D array for storing cost rate function
    % return values.
    l_mat = zeros(n1, n2, nu);
    
    
    % Iterate through all (x1, x2, u) tuples and fill y1/y2/l matrices
    for x1_idx=1:n1,
        for x2_idx=1:n2,
            for u_idx=1:nu,
                x1 = x1_vals(x1_idx);
                x2 = x2_vals(x2_idx);
                u = u_vals(u_idx);
                
                % Calculate next state via system dynamics
                y1_val = x1 + h * x2;
                y2_val = x2 + h*(a*sin(x1) - b*x2 + u);
                
                % Store the neighboring states and their
                % respective probabilities in structs
                y1_neighbors = get_neighbors(y1_val, x1_vals, 0);
                y2_neighbors = get_neighbors(y2_val, x2_vals, 1);

                % Store input index that produces this transition, for
                % future reference
                y1_neighbors.input = u_idx;
                y2_neighbors.input = u_idx;
                
                % Store structs in y matrices
                y1_mat(x1_idx, x2_idx, u_idx) = y1_neighbors;
                y2_mat(x1_idx, x2_idx, u_idx) = y2_neighbors;
                
                % Store cost rates in l matrix
                l_mat(x1_idx, x2_idx, u_idx) = 1-exp(k*cos(x1)-k)+(r/2)*u*u;
            end
        end
    end
end