function mdp_solver()
    % Parameter values
    % h          time step
    % v_max      maximum velocity
    % u_max      maximum control magnitude
    % n1         # of grid points on state x1 (arc) dimension
    % n2         # of grid points on state x2 (velocity) dimension
    % nu         # of grid points on u (control) dimension
    % a          models gravity, mass, and length ( > 0)
    % b          damping factor ( > 0)
    % k          sharpness of cost function ( > 0 ); penalty for being not
    %               at peak of circle
    % r          weight of control cost ( > 0 )    
    h = 0.025;
    v_max = 10;
    u_max = 8;
    n1 = 17;
    n2 = 27;
    nu = 15;
    a = 0.15;
    b = 0.5;
    k = 0.7;
    r = 0.0001;

    x1_vals = linspace(0, 2*pi, n1); % divide full arc into n1 pts
    x2_vals = linspace(-v_max, v_max, n2); % divide velocity into n2 pts
    u_vals = linspace(-u_max, u_max, nu); % divide velocity into nu pts
    
    
    % Make the MDP
    [y1_mat, y2_mat, l_mat] = construct_mdp(h, v_max, u_max, n1, n2, nu, a, b, k, r);

    % value iteration
    alpha = 0.5; % discount factor
    conv_threshold = 1E-10; % convergence threshold
    cur_values = zeros(n1,n2); % value function data
    cur_policies = ones(n1,n2,1)*round(nu+1)/2;      % policy    
    hamiltonians = zeros(n1,n2,nu); % hamiltonian function data
    
    while 1,
        disp('iterating');
        old_values = cur_values;

        % compute Hamiltonian for current values, based on cost rate and
        % transition probabilities. This super-deep nested for loop
        % iterates through all inputs (u_idx), & states (x1_idx and x2_idx)
        for u_idx = 1:nu,
            for x1_idx = 1:n1,
                for x2_idx = 1:n2,
                    % Sum probability-weighted values for each
                    % action, based on what next states they can
                    % take you to
                    next1.points = y1_mat(x1_idx,x2_idx,u_idx).points;
                    next1.indices = y1_mat(x1_idx,x2_idx,u_idx).indices;
                    next1.probs = y1_mat(x1_idx,x2_idx,u_idx).probs;
                    next2.points = y2_mat(x1_idx,x2_idx,u_idx).points;
                    next2.indices = y2_mat(x1_idx,x2_idx,u_idx).indices;
                    next2.probs = y2_mat(x1_idx,x2_idx,u_idx).probs;
                    
                    estimated_value_sum = 0;
                    for n1_i = 1:length(next1.indices),
                        for n2_i = 1:length(next2.indices),
                            this_state_value = cur_values(next1.indices(n1_i), next2.indices(n2_i));
                            this_state_prob = next1.probs(n1_i) * next2.probs(n2_i);
                            estimated_value_sum = estimated_value_sum + this_state_value * this_state_prob;
                        end
                    end
                    % Now that we have the estimated value of this
                    % state/input tuple, calculate the Hamiltonian
                    hamiltonians(x1_idx,x2_idx,u_idx) = l_mat(x1_idx,x2_idx,u_idx) + alpha*estimated_value_sum;
                end
            end
        end

        % update value, compute policy
        [cur_values, cur_policies] = min(hamiltonians,[],3);
        
        % Remove singleton dimension
        cur_values = squeeze(cur_values);
        
        % check for convergence
        if max(abs(cur_values-old_values))<conv_threshold,
            break;
        end
    end

    opt_policies = cur_policies;
    opt_values = cur_values;
    
    % Plot cost-to-go
   
%     surf(x1_vals, x2_vals, opt_values');
%     xlabel('x1 (Position)');
%     ylabel('x2 (Velocity)');
%     zlabel('Cost-To-Go');
%     
%     return    
    
    
    % Plot optimal control
%     opt_policy_values = zeros(size(opt_policies));
%     for i = 1:n1,
%         for j = 1:n2,
%             opt_policy_values(i,j) = u_vals(opt_policies(i,j));
%         end
%     end
%    
%     surf(x1_vals, x2_vals, opt_policy_values');
%     xlabel('x1 (Position)');
%     ylabel('x2 (Velocity)');
%     zlabel('Optimal Control');
%     
%     return
    
    
    disp('Policies calculated; running simulation...');
    
    % Run simulation using policies, starting with 0 ICs
    x1_idx = find(x1_vals==pi, 1); % radian position measured from top, so we start at pi radians
    x2_idx = find(x2_vals==0, 1);

    iterationCount = 1;
    for i=1:500,
        % First, draw the current state (animated!)
        clf;
        figure(1)
        filename = 'pendulum.gif';
        draw_state(x1_vals(x1_idx), filename, iterationCount);
        iterationCount = iterationCount + 1;
        

%          disp('n:');
%          disp(iterationCount);
%          disp('x1:');
%          disp(x1_vals(x1_idx));
%         disp('x2:');
%         disp(x2_vals(x2_idx));
%         
        
        % Get input that policy recommends in this state
         u_idx = opt_policies(x1_idx, x2_idx);

%         disp('u from policy:');
%         disp(u_vals(u_idx));
%         
        
        % Figure out all of the possible next states and their
        % probabilities
        next1.indices = y1_mat(x1_idx,x2_idx,u_idx).indices;
        next1.probs = y1_mat(x1_idx,x2_idx,u_idx).probs;
        next2.indices = y2_mat(x1_idx,x2_idx,u_idx).indices;
        next2.probs = y2_mat(x1_idx,x2_idx,u_idx).probs;

        if length(next1.indices) == 1,
            % Go with that next state; it's the only possible one
            x1_idx = next1.indices(1);
        else
            % Do a dice roll to see what the next state will be
            x1_idx = randsample(next1.indices, 1, true, next1.probs);
        end
        
        if length(next1.indices) == 1,
            % Go with that next state; it's the only possible one
            x2_idx = next2.indices(1);
        else
            % Do a dice roll to see what the next state will be
            x2_idx = randsample(next2.indices, 1, true, next2.probs);
        end
        
        pause(0.015);
    end
    
    
end